from data_providers.jira.issue_provider import JiraIssueProvider
from data_providers.jira.story_point_extractor import JiraCustomFieldStoryPointExtractor, \
    JiraTShirtStoryPointExtractor
from data_providers.jira.worklog_extractor import JiraWorklogExtractor, JiraStatusChangeWorklogExtractor



from datetime import datetime
from typing import Dict

from data_providers import WorklogExtractor
from data_providers.worklog_extractor import IssueTotalSpentTimeExtractor
from data_providers.worktime_extractor import SimpleWorkTimeExtractor, WorkTimeExtractor


class JiraWorklogExtractor(WorklogExtractor):

    def __init__(self, jira_client, user_filter: list[str] = None, include_subtask_worklog=False) -> None:
        self.jira_client = jira_client
        self.user_filter = user_filter
        self.include_subtask_worklog = include_subtask_worklog

    def get_work_time_per_user(self, issue):
        worklogs = self._get_worklog_for_issue_with_subtasks(issue)

        working_time_per_user = {}
        for worklog in worklogs:
            worklog_user = self._extract_user_from_worklog(worklog)
            if self._is_allowed_user(worklog_user):
                worklog_time_spent = self._extract_time_in_seconds_from_worklog(worklog)

                working_time_per_user[worklog_user] = working_time_per_user.get(worklog_user, 0) + worklog_time_spent

        return working_time_per_user

    def _get_worklog_for_issue_with_subtasks(self, issue):
        worklogs = []
        worklogs.extend(self._get_worklogs_from_jira(issue['key']))
        if self.include_subtask_worklog:
            try:
                sub_task_keys = [subtask["key"] for subtask in issue["fields"]["subtasks"]]
                for subtask in sub_task_keys:
                    worklogs.extend(self._get_worklogs_from_jira(subtask))
            except AttributeError:
                pass
        return worklogs

    def _get_worklogs_from_jira(self, issue_key: str):
        data = self.jira_client.issue_get_worklog(issue_key)
        if 'worklogs' in data:
            return data['worklogs']
        return data

    def _is_allowed_user(self, worklog_user):
        if self.user_filter is None:
            return True
        if worklog_user in self.user_filter:
            return True
        return False

    @staticmethod
    def _extract_time_in_seconds_from_worklog(worklog):
        return worklog["timeSpentSeconds"]

    @staticmethod
    def _extract_user_from_worklog(worklog):
        return worklog["author"]["accountId"]


SIMPLE_WORKTIME_EXTRACTOR = SimpleWorkTimeExtractor()


class JiraStatusChangeWorklogExtractor(WorklogExtractor):

    def __init__(self, transition_statuses: list[str],
                 user_filter: list[str] = None,
                 time_format='%Y-%m-%dT%H:%M:%S.%f%z',
                 use_user_name=False,
                 use_status_codes=False,
                 worktime_extractor: WorkTimeExtractor = SIMPLE_WORKTIME_EXTRACTOR) -> None:

        self.transition_status_codes = transition_statuses
        self.use_status_codes = use_status_codes
        self.user_filter = user_filter

        self.time_format = time_format

        self.worktime_extractor = worktime_extractor

        self.use_user_name = use_user_name
        self.interval_start_time = None
        self.interval_end_time = None

    def get_work_time_per_user(self, issue) -> Dict[str, int]:
        working_time_per_user = {}

        changelog_history = self._extract_issue_changelog_history(issue)
        if len(changelog_history) == 0:
            return working_time_per_user

        last_assigned_user = self._default_assigned_user()
        for changelog_entry in changelog_history:
            if self.__is_user_change_entry(changelog_entry):
                assignee = self._extract_user_from_changelog(changelog_entry)
                if self._is_allowed_user(assignee):
                    last_assigned_user = assignee
            elif self._is_status_change_entry(changelog_entry):
                if self._is_status_changed_into_required(changelog_entry):
                    if self.interval_start_time is None:
                        self.interval_start_time = datetime.strptime(changelog_entry['created'], self.time_format)
                elif self._is_status_changed_from_required(changelog_entry):
                    if self.interval_end_time is None:
                        self.interval_end_time = datetime.strptime(changelog_entry['created'], self.time_format)

                self.__sum_working_time(working_time_per_user, last_assigned_user)

        if self._is_current_status_a_required_status(issue):
            self.interval_end_time = datetime.now().astimezone()
            self.__sum_working_time(working_time_per_user, last_assigned_user)

        return working_time_per_user

    def __sum_working_time(self, working_time_per_user, last_assigned_user):
        if self.__is_interval_found_for_status_change():
            seconds_in_status = self.worktime_extractor.extract_time_from_period(self.interval_start_time,
                                                                                 self.interval_end_time)
            if seconds_in_status is not None:
                working_time_per_user[last_assigned_user] = working_time_per_user.get(
                    last_assigned_user, 0) + seconds_in_status

            self.__clean_interval_times()

    def _extract_issue_changelog_history(self, issue):
        """ Create a flat list of user or status changelog history entries
        and add created time into each history entry """

        if not isinstance(issue, dict):
            return []

        if 'changelog' not in issue or 'histories' not in issue['changelog']:
            return []

        changelog_history = []
        changelog = issue['changelog']['histories']
        for history_entry in changelog:
            if 'items' not in history_entry:
                continue
            for history_entry_item in history_entry['items']:
                if self._is_status_change_entry(history_entry_item) or self.__is_user_change_entry(history_entry_item):
                    history_entry_item['created'] = history_entry['created']
                    changelog_history.append(history_entry_item)

        # Jira API sends changelog in last entry first,
        # so we want to reverse it to iterate for better iteration
        changelog_history.reverse()
        return changelog_history

    def _extract_user_from_changelog(self, changelog_entry):
        if self.use_user_name:
            return changelog_entry['toString']
        else:
            return changelog_entry['to']

    def _is_allowed_user(self, last_assigned):
        if self.user_filter is None:
            return True
        if last_assigned in self.user_filter:
            return True
        return False

    @staticmethod
    def _is_status_change_entry(changelog_entry):
        return 'fieldId' in changelog_entry and changelog_entry['fieldId'] == 'status'

    @staticmethod
    def __is_user_change_entry(changelog_entry):
        return 'fieldId' in changelog_entry and changelog_entry['fieldId'] == 'assignee'

    @staticmethod
    def _default_assigned_user():
        return 'UNKNOWN'

    def _is_status_changed_into_required(self, changelog_entry):
        if self.transition_status_codes is None:
            return True
        if self.use_status_codes:
            return changelog_entry['to'] in self.transition_status_codes
        else:
            return changelog_entry['toString'] in self.transition_status_codes

    def _is_status_changed_from_required(self, changelog_entry):
        if self.transition_status_codes is None:
            return True
        if self.use_status_codes:
            return changelog_entry['from'] in self.transition_status_codes
        else:
            return changelog_entry['fromString'] in self.transition_status_codes

    def _is_current_status_a_required_status(self, issue):
        if self.transition_status_codes is None:
            return True
        if 'fields' not in issue or 'status' not in issue['fields']:
            return False
        if self.use_status_codes:
            return issue['fields']['status']['id'] in self.transition_status_codes in self.transition_status_codes
        else:
            return issue['fields']['status']['name'] in self.transition_status_codes

    def __is_interval_found_for_status_change(self):
        return self.interval_start_time is not None and self.interval_end_time is not None

    def __clean_interval_times(self):
        self.interval_start_time = None
        self.interval_end_time = None


class JiraResolutionTimeIssueTotalSpentTimeExtractor(IssueTotalSpentTimeExtractor):

    def __init__(self, time_format='%Y-%m-%dT%H:%M:%S.%f%z', ) -> None:
        self.time_format = time_format

    def get_total_spent_time(self, issue) -> int:
        resolution_date_str = issue['fields']['resolutiondate']
        if resolution_date_str is None:
            return 0

        resolution_date = datetime.strptime(resolution_date_str, self.time_format)
        creation_date = datetime.strptime(issue['fields']['created'], self.time_format)
        spent_time = (resolution_date - creation_date)
        return int(spent_time.total_seconds())



from abc import abstractmethod, ABC


class IssueProvider(ABC):

    @abstractmethod
    def get_issues(self) -> list:
        pass


class ProxyIssueProvider(IssueProvider):

    def __init__(self, issues: list) -> None:
        self.issues = issues

    def get_issues(self) -> list:
        return self.issues



from abc import ABC, abstractmethod
from typing import Dict

from calculators.metrics_calculator import MetricCalculator
from calculators.utils import time_utils
from data_providers import WorklogExtractor
from data_providers.issue_provider import IssueProvider
from data_providers.story_point_extractor import StoryPointExtractor
from data_providers.utils import VelocityTimeUnit
from data_providers.worklog_extractor import IssueTotalSpentTimeExtractor


class JiraMetricsCalculator(MetricCalculator, ABC):

    def __init__(self) -> None:
        self.data_fetched = False

    def calculate(self, velocity_time_unit=VelocityTimeUnit.DAY) -> Dict[str, float]:
        if not self.is_data_fetched():
            self._extract_data_from_issues()
            self.mark_data_fetched()
        self._calculate_metric(velocity_time_unit)
        return self.get_metric()

    def mark_data_fetched(self):
        self.data_fetched = True

    def is_data_fetched(self):
        return self.is_data_fetched is True

    @abstractmethod
    def _calculate_metric(self, time_unit: VelocityTimeUnit):
        pass

    @abstractmethod
    def _extract_data_from_issues(self):
        pass

    @abstractmethod
    def get_metric(self) -> Dict[str, float]:
        pass


class UserVelocityCalculator(JiraMetricsCalculator):

    def __init__(self, issue_provider: IssueProvider,
                 story_point_extractor: StoryPointExtractor,
                 worklog_extractor: WorklogExtractor) -> None:
        super().__init__()
        self.issue_provider = issue_provider
        self.story_point_extractor = story_point_extractor
        self.worklog_extractor = worklog_extractor

        self.velocity_per_user = {}
        self.resolved_story_points_per_user = {}
        self.time_in_seconds_spent_per_user = {}

    def _calculate_metric(self, time_unit: VelocityTimeUnit):
        for user in self.resolved_story_points_per_user:
            spent_time_in_seconds = self.time_in_seconds_spent_per_user[user]
            if spent_time_in_seconds != 0:
                spent_time = time_utils.convert_time(spent_time_in_seconds, time_unit)
                developer_velocity = self.resolved_story_points_per_user[user] / spent_time
                if developer_velocity != 0:
                    self.velocity_per_user[user] = developer_velocity

    def _extract_data_from_issues(self):
        issues = self.issue_provider.get_issues()
        for issue in issues:
            issue_story_points = self.story_point_extractor.get_story_points(issue)
            if issue_story_points is not None and issue_story_points > 0:
                time_user_worked_on_issue = self.worklog_extractor.get_work_time_per_user(issue)

                self._sum_story_points_and_worklog(issue_story_points, time_user_worked_on_issue)

    def get_metric(self):
        return self.velocity_per_user

    def get_story_points(self):
        return self.resolved_story_points_per_user

    def get_spent_time(self):
        return self.time_in_seconds_spent_per_user

    def _sum_story_points_and_worklog(self, issue_story_points, time_user_worked_on_issue):
        issue_total_spent_time = float(sum(time_user_worked_on_issue.values()))
        if issue_total_spent_time == 0:
            return

        for user in time_user_worked_on_issue.keys():
            if user not in self.resolved_story_points_per_user:
                self.resolved_story_points_per_user[user] = 0.
            if user not in self.time_in_seconds_spent_per_user:
                self.time_in_seconds_spent_per_user[user] = 0

        for user in time_user_worked_on_issue.keys():
            story_point_ratio = time_user_worked_on_issue[user] / issue_total_spent_time
            self.resolved_story_points_per_user[user] += issue_story_points * story_point_ratio
            self.time_in_seconds_spent_per_user[user] += time_user_worked_on_issue[user]


class GeneralizedTeamVelocityCalculator(JiraMetricsCalculator):

    def __init__(self, issue_provider: IssueProvider,
                 story_point_extractor: StoryPointExtractor,
                 time_extractor: IssueTotalSpentTimeExtractor) -> None:
        super().__init__()
        self.total_resolved_story_points = 0
        self.total_spent_time_in_seconds = 0
        self.velocity = None

        self.issue_provider = issue_provider
        self.story_point_extractor = story_point_extractor
        self.time_extractor = time_extractor

    def _calculate_metric(self, time_unit: VelocityTimeUnit):
        spent_time = time_utils.convert_time(self.total_spent_time_in_seconds, time_unit)
        story_points = self.total_resolved_story_points

        if spent_time == 0:
            self.velocity = 0
        else:
            self.velocity = story_points / spent_time

    def _extract_data_from_issues(self):
        issues = self.issue_provider.get_issues()
        for issue in issues:
            issue_story_points = self.story_point_extractor.get_story_points(issue)
            if issue_story_points is not None and issue_story_points > 0:
                time_spent_on_issue = self.time_extractor.get_total_spent_time(issue)

                self._sum_story_points_and_worklog(issue_story_points, time_spent_on_issue)

    def get_metric(self) -> Dict[str, float]:
        return self.velocity

    def get_story_points(self):
        return self.total_resolved_story_points

    def get_spent_time(self):
        return self.total_spent_time_in_seconds

    def _sum_story_points_and_worklog(self, issue_story_points: float, issue_total_spent_time: int):
        if issue_total_spent_time == 0:
            return

        self.total_resolved_story_points += issue_story_points
        self.total_spent_time_in_seconds += issue_total_spent_time



import calendar
import datetime
from enum import Enum, auto

from dateutil.relativedelta import relativedelta

from data_providers.utils import VelocityTimeUnit


class JiraIssueSearchQueryBuilder:
    class __QueryParts(Enum):
        PROJECT = auto()
        TYPE = auto()
        STATUS = auto()
        RESOLUTION_DATE = auto()
        LAST_MODIFIED = auto()

    def __init__(self,
                 projects: list[str] = None,
                 resolution_dates: tuple[datetime.datetime, datetime.datetime] = None,
                 statuses: list[str] = None,
                 issue_types: list[str] = None
                 ) -> None:
        self.query_parts = {}

        self.with_projects(projects)
        self.with_statuses(statuses)
        self.for_resolution_dates(resolution_dates)
        self.for_issue_types(issue_types)

    def with_projects(self, projects: list[str]):
        if projects is None:
            return
        project_filter = "project IN (" + ",".join(projects) + ")"
        self.__add_filter(self.__QueryParts.PROJECT, project_filter)

    def with_statuses(self, statuses):
        if statuses is None:
            return
        status_filter = "status in (" + self.__convert_in_jql_value_list(statuses) + ")"
        self.__add_filter(self.__QueryParts.STATUS, status_filter)

    def for_resolution_dates(self, resolution_dates: tuple[datetime.datetime, datetime.datetime]):
        if resolution_dates is None:
            return
        date_filter = self.__create_date_range_filter("resolutiondate",
                                                      resolution_dates[0],
                                                      resolution_dates[1])
        self.__add_filter(self.__QueryParts.RESOLUTION_DATE, date_filter)

    def for_last_modified_dates(self, last_modified_datas: tuple[datetime.datetime, datetime.datetime]):
        if last_modified_datas is None:
            return
        date_filter = self.__create_date_range_filter("updated",
                                                      last_modified_datas[0],
                                                      last_modified_datas[1])
        self.__add_filter(self.__QueryParts.LAST_MODIFIED, date_filter)

    def for_issue_types(self, issue_types):
        if issue_types is None:
            return
        issue_type_filter = "issuetype in (" + self.__convert_in_jql_value_list(issue_types) + ")"
        self.__add_filter(self.__QueryParts.TYPE, issue_type_filter)

    def build_query(self) -> str:
        return ' AND '.join(self.query_parts.values())

    @staticmethod
    def __convert_in_jql_value_list(statuses):
        return ', '.join(['"%s"' % w for w in statuses])

    @staticmethod
    def __create_date_range_filter(field_name: str, start_date: datetime.date, end_date: datetime.date):
        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d')
        return "%s >= '%s' and %s <= '%s'" % (field_name, start_date_str, field_name, end_date_str)

    def __add_filter(self, query_part_type: __QueryParts, query_part):
        self.query_parts[query_part_type] = query_part.strip()


class TimeRangeGenerator:

    def __init__(self, time_unit: VelocityTimeUnit, number_of_ranges: int) -> None:
        self.time_unit = time_unit
        self.number_of_ranges = number_of_ranges
        self.period_initial_date = datetime.datetime.today()

    def __iter__(self):
        for i in range(self.number_of_ranges):
            yield tuple((self.__resolve_start_date_of_period(), self.__resolve_end_date_of_period()))
            self.__decrease_date_range()

    def __decrease_date_range(self):
        if self.time_unit == VelocityTimeUnit.HOUR:
            self.period_initial_date -= relativedelta(hours=1)
        elif self.time_unit == VelocityTimeUnit.DAY:
            self.period_initial_date -= relativedelta(days=1)
        elif self.time_unit == VelocityTimeUnit.WEEK:
            self.period_initial_date -= relativedelta(weeks=1)
        elif self.time_unit == VelocityTimeUnit.MONTH:
            self.period_initial_date -= relativedelta(months=1)

    def __resolve_start_date_of_period(self) -> datetime.datetime:
        if self.time_unit == VelocityTimeUnit.HOUR:
            return self.period_initial_date.replace(minute=0, second=0, microsecond=0)
        elif self.time_unit == VelocityTimeUnit.DAY:
            return self.period_initial_date.replace(hour=0, minute=0, second=0, microsecond=0)
        elif self.time_unit == VelocityTimeUnit.WEEK:
            current_day_of_week = self.period_initial_date.weekday()
            delta_with_first_day_of_week = datetime.timedelta(days=current_day_of_week)
            return self.period_initial_date - delta_with_first_day_of_week
        elif self.time_unit == VelocityTimeUnit.MONTH:
            return self.period_initial_date.replace(day=1)

    def __resolve_end_date_of_period(self):
        if self.time_unit == VelocityTimeUnit.HOUR:
            return self.period_initial_date.replace(minute=59, second=59, microsecond=999999)
        elif self.time_unit == VelocityTimeUnit.DAY:
            return self.period_initial_date.replace(hour=23, minute=59, second=59, microsecond=999999)
        elif self.time_unit == VelocityTimeUnit.WEEK:
            current_day_of_week = self.period_initial_date.weekday()
            delta_with_last_day_of_week = datetime.timedelta(days=6 - current_day_of_week)
            return self.period_initial_date - delta_with_last_day_of_week
        elif self.time_unit == VelocityTimeUnit.MONTH:
            last_day_of_month = calendar.monthrange(self.period_initial_date.year, self.period_initial_date.month)[1]
            return self.period_initial_date.replace(day=last_day_of_month)



import concurrent
import math
from concurrent.futures import ThreadPoolExecutor, wait

from data_providers import IssueProvider


class JiraIssueProvider(IssueProvider):

    def __init__(self,
                 jira_client,
                 query: str,
                 expand: [str] = None,
                 thread_pool_executor: ThreadPoolExecutor = None) -> None:
        self.jira_client = jira_client
        self.query = query.strip()
        self.expand = expand
        if expand is None:
            self.expand_str = None
        else:
            self.expand_str = ",".join(self.expand)
        self.thread_pool_executor = thread_pool_executor

    def get_issues(self):
        first_page = self.jira_client.jql(self.query, expand=self.expand_str, limit=self._get_issue_fetch_amount())
        first_page_issues = first_page["issues"]
        issues_total_count = first_page["total"]
        page_len = len(first_page_issues)

        issues = []
        issues.extend(first_page_issues)

        if page_len < issues_total_count:
            amount_of_fetches = math.ceil(issues_total_count / float(page_len))

            if self.thread_pool_executor is None:
                self._fetch_issue_sync(issues, amount_of_fetches, page_len)
            else:
                self._fetch_issue_concurrently(issues, amount_of_fetches, page_len)

        return issues

    def _fetch_issue_concurrently(self, issues, amount_of_fetches, page_len):
        features = []
        for i in range(1, amount_of_fetches):
            next_search_start = i * page_len
            feature = self.thread_pool_executor.submit(self.jira_client.jql,
                                                       self.query,
                                                       expand=self.expand_str,
                                                       limit=self._get_issue_fetch_amount(),
                                                       start=next_search_start)
            features.append(feature)
        done, not_done = wait(features, return_when=concurrent.futures.ALL_COMPLETED)
        for feature in done:
            issues.extend(feature.result()["issues"])

    def _fetch_issue_sync(self, issues, amount_of_fetches, page_len):
        for i1 in range(1, amount_of_fetches):
            start = i1 * page_len
            current_page_result = self.jira_client.jql(self.query,
                                                       expand=self.expand_str,
                                                       limit=self._get_issue_fetch_amount(),
                                                       start=start)
            current_page_issues = current_page_result["issues"]
            issues.extend(current_page_issues)

    def _get_issue_fetch_amount(self):
        if self.thread_pool_executor is None:
            return 100
        else:
            return 50


class CachingJiraIssueProvider(JiraIssueProvider):

    def __init__(self,
                 jira_client,
                 query: str,
                 expand: [str] = None,
                 thread_pool_executor: ThreadPoolExecutor = None,
                 cache=None) -> None:
        super().__init__(jira_client, query, expand, thread_pool_executor)
        self.cache = cache

    def get_issues(self):
        # JIRA API is broken and set of maxResults doesn`t works.
        # return jira.jql(jql_query, limit=MAX_SEARCH_RESULTS)

        cached_issues = self._search_in_cache()
        if cached_issues is not None:
            return cached_issues

        issues = super(CachingJiraIssueProvider, self).get_issues()

        self._set_in_cache(issues)

        return issues

    def _search_in_cache(self):
        if self.cache is None:
            return None

        # if query with same expands is already present in cache
        cache_key = self._create_cache_key_for_query()
        if self._is_key_in_cache(cache_key):
            return self._get_from_cache(cache_key)

        # searching for cached responses with not less expands
        # for example we have in cache query with worklog expand and searching
        # fir same query without expands. In such case we can safely return cached results.
        for key in self._get_all_cache_keys():
            if key.startswith(self.query):
                if self.expand is None:
                    return self._get_from_cache(key)
                else:
                    all_expands_present = True
                    for expand in self.expand:
                        if "_" + expand not in key:
                            all_expands_present = False
                    if all_expands_present:
                        return self._get_from_cache(key)

        return None

    def _get_from_cache(self, cache_key):
        return self.cache.get(cache_key)

    def _set_in_cache(self, issues):
        if self.cache is None:
            return
        self.cache[self._create_cache_key_for_query()] = issues

    def _is_key_in_cache(self, cache_key):
        return cache_key in self.cache

    def _get_all_cache_keys(self):
        return self.cache.keys()

    def _create_cache_key_for_query(self):
        if self.expand is None:
            return self.query
        return self.query + "_" + "_".join(self.expand)
